program SPELDX;

const
  version         = '2.04';
  dictionary_file = 'DX.DAT';
  errorfile       = 'MISSPELL.DAT';
  dxsize          = 3855;
  error_file_size = 1000;
  wordlen         = 16;

type
  wordtype=string[16];
  dxtype   = array[1..dxsize] of string[wordlen];
  dxptr    = ^dxtype;

var
  found,
  yes         :boolean;
  choice,
  suffix1     :char;
  suffix2     :string[2];
  i,
  num_entries,
  size,
  word_no     : integer;
  dxfile,
  mispfile    : text;
  entry       : dxptr;
  list        : array [0..error_file_size] of string[wordlen];
  word        : string[wordlen];

procedure dashline;
  begin
   writeln('==============================================================================');
  end;

procedure beep;
begin
  write(chr(7));
end;

procedure query(bias:boolean);
var
  yesno       :char;
begin
  yes:=bias;
  readln(yesno);
  if yesno in ['Y','y'] then yes:=true;
  if yesno in ['N','n'] then yes:=false;
end;

procedure waitkey;
begin
  writeln;
  write('* Press any key to continue ');
  repeat until keypressed;
end;

procedure read_file;
begin
  assign(dxfile,dictionary_file);
  reset(dxfile);
  i:=0;
  while not eof(dxfile) do
    begin
      i:=i+1;
      readln(dxfile,entry^[i]);
    end;
  close(dxfile);
  num_entries:=i;
end;

procedure write_file;
begin
  assign(dxfile,dictionary_file);
  rewrite(dxfile);
  for i:=1 to num_entries do
    begin
      write(dxfile,entry^[i]);
      writeln(dxfile);
    end;
  close(dxfile);
end;

procedure bsearch;

var
  lower,
  upper,
  center : integer;

begin
  lower:=1;
  upper:=num_entries;
  found:=false;
  while (upper>=lower) and (not found) do
  begin
    center:=(lower+upper) div 2;
    if word=entry^[center] then
      begin
        found:=true;
      end
    else
      if word>entry^[center] then
        lower:=center+1
      else
        upper:=center-1
  end;
  word_no:=center;
end;

procedure shell_sort;
var
  done  :boolean;
  jump,
  i,
  j,
  swno  :integer;

function firstjump(length: integer): integer;
var
  temp : integer;
begin
  temp:=1;
  while temp<=length do
    temp:=temp*2;
  firstjump:=temp
end;

procedure swap (var p,q: wordtype);
var
  hold :wordtype;
begin
  hold:=p;
  p:=q;
  q:=hold
end;

begin
  swno:=0;
  jump:=firstjump(num_entries);
  writeln;write('* Sorting ');
  while jump>1 do
  begin
    jump:=(jump-1) div 2;
    write('.');
      repeat
        done:=true;
        for j:=1 to num_entries-jump do
        begin
          i:=j+jump;
          if entry^[j]>entry^[i] then
          begin
            swno:=swno+1;
            swap(entry^[j],entry^[i]);
            done:=false;
          end;
        end;
      until done
  end;
end;

procedure kill_word;
var
  i  : integer;
begin
  bsearch;
  if not found then
    begin
      writeln;writeln('  "',word,'" not found');
      beep;
    end
  else
    begin
      writeln;writeln('  Deleting "',word,'"');
      for i:=word_no+1 to num_entries do
        entry^[i-1]:=entry^[i];
      num_entries:=num_entries-1;
    end;
end;

procedure lowcase;
var
  ch        :char;
  j         :integer;
  nuword    :string[wordlen];
begin
  nuword:='';
  for j:=1 to length(word) do
    begin
      ch:=copy(word,j,1);
      if ch in ['A'..'Z'] then ch:=chr(ord(ch)+32);
      nuword:=nuword+ch;
    end;
  word:=nuword;
end;

procedure max_warning;
begin
  if num_entries=dxsize-10 then
    begin
      beep;writeln;
      writeln('  WARNING - YOU ARE WITHIN 10 WORDS OF DICTIONARY SIZE LIMIT !');
    end;
  if num_entries=dxsize then
    begin
      beep;writeln;
      writeln('* DICTIONARY SIZE EXCEEDED -- NO MORE WORDS MAY BE ADDED !');
      yes:=false;
    end;
end;

procedure add_list;
var
  i,j,
  l,
  k      : integer;
begin
  clrscr;
  writeln('* Add Words from Misspelled File');
  writeln;writeln('  Entries in Dictionary > ',num_entries);
  writeln;write('* Reading Misspelled File > ');
  assign(mispfile,errorfile);
  reset(mispfile);
  j:=1;
  while not eof(mispfile) do
    begin
      readln(mispfile,list[j]);
      j:=j+1;
    end;
  close(mispfile);
  size:=j;
  writeln(size-1,' entries');
  i:=num_entries;
  k:=1;
  repeat
    begin
      repeat
        begin
          i:=i+1;
          writeln;
          writeln('* "',list[k],'" -- Enter in Correct Form Below');writeln;
          write('  Enter (Press RETURN to skip) > ');
          readln(word);lowcase;
          writeln;write('  Correct (Y/N)? <Y> ');
          query(true);
          if yes then entry^[i]:=word
          else i:=i-1;
        end;
      until yes;
      bsearch;
      if found then
        begin
          writeln;writeln('  "',word,'" already in dictionary');
          beep;
          i:=i-1;
        end;
      num_entries:=i;
      max_warning;
      writeln;write('  Continue (Y/N)? <Y> ');
      query(true);
    end;
    k:=k+1;
    if k=size then
      begin
        yes:=false;
        beep;writeln;
        writeln('* END OF "MISSPELLED" FILE');
      end;
  until not yes;
  num_entries:=i;
  shell_sort;
  writeln;writeln;writeln('* Checking for Duplicate Words ');writeln;
  k:=2;
  while k<=num_entries do
    begin
      if entry^[k]=entry^[k-1] then
        begin
          writeln;writeln('  Deleting "',entry^[k],'" as duplicate');
          for l:=k to num_entries do
            entry^[l]:=entry^[l+1];
          k:=k-1;
          num_entries:=num_entries-1;
        end;
      k:=k+1;
    end;
  word:='';
  bsearch;
  if found then kill_word;
  waitkey;
end;

procedure add_entry;
var
  i,
  l,
  k      : integer;
  nuword :string[16];
begin
  clrscr;
  writeln('* Add Entry');
  writeln;writeln('  Entries in Dictionary > ',num_entries);
  i:=num_entries;
  repeat
    begin
      repeat
        begin
          i:=i+1;
          writeln;
          write('* Enter New Word > ');
          readln(word);lowcase;
          writeln;write('  Correct (Y/N)? <Y> ');
          query(true);
          if yes then entry^[i]:=word
          else i:=i-1;
        end;
      until yes;
      bsearch;
      if found then
        begin
          writeln;writeln('  "',word,'" already in dictionary');
          beep;
          i:=i-1;
        end;
      num_entries:=i;
      writeln;write('  Continue (Y/N)? <Y> ');
      query(true);
      max_warning;
    end;
  until not yes;
  num_entries:=i;
  shell_sort;
  writeln;writeln;writeln('* Checking for Duplicate Words ');writeln;
  k:=2;
  while k<=num_entries do
    begin
      if entry^[k]=entry^[k-1] then
        begin
          writeln;writeln('  Deleting "',entry^[k],'" as duplicate');
          for l:=k to num_entries do
            entry^[l]:=entry^[l+1];
          k:=k-1;
          num_entries:=num_entries-1;
        end;
      k:=k+1;
    end;
  word:='';
  bsearch;
  if found then kill_word;
  waitkey;
end;

procedure delete;
begin
  clrscr;writeln;
  repeat
    writeln;write('* Delete Word > ');
    readln(word);
    kill_word;
    writeln;write('  Continue (Y/N)? <Y> ');
    query(true);
  until not yes;
end;

procedure find;
var
  bottom,
  i,
  top       : integer;
begin
  clrscr;writeln;
  repeat
    found:=false;
    write('* Find Word > ');
    read(word);
    bsearch;
    if found then
      begin
        writeln;writeln;
        writeln('  "',word,'" in dictionary as entry # ',word_no,' out of ',num_entries);
      end
    else
      begin
        writeln('   NOT IN DICTIONARY');
        writeln;write('  Closest > ');
        bottom:=word_no-2;if bottom<=0 then bottom:=1;
        top:=word_no+2; if top>=num_entries then top:=num_entries;
        for i:=bottom to top do
          write(entry^[i],'  ');
        writeln;
        beep;
      end;
    writeln;write('  Continue (Y/N)? <Y> ');
    query(true);writeln;
  until not yes;
end;

procedure display;
var
  i : integer;
begin
  writeln;writeln('* Display Dictionary');
  writeln;write('  Begin Display with Letter(s)  (def.="a") > ');
  readln(word);
  clrscr;
  bsearch;
  i:=word_no;
  yes:=true;
  while yes and (i<=num_entries) do
    begin
      writeln(i:2,'   ',entry^[i]);
      if i mod 22=0 then
        begin
          writeln;write('Continue ? (Y/N) > ');
          query(true);
          writeln;writeln;
        end;
      i:=i+1;
    end;
  if i>num_entries-22 then waitkey;
end;

procedure size_of_dx;
begin
  writeln;writeln;writeln('* Size of Dictionary > ',num_entries,' entries');
  waitkey;
end;

procedure redundant;
var
  i     : integer;
begin
  clrscr;
  writeln('* Removing Redundant Words ');
  for i:=1 to num_entries do
  begin
    suffix1:=copy(entry^[i],length(entry^[i]),1);
    if suffix1 in ['d','s'] then
      begin
        word:=copy(entry^[i],1,length(entry^[i])-1);
        bsearch;
        if found then
          begin
            writeln;writeln('  ',word,'  ',entry^[i]);
            writeln;write('  Remove (Y/N) ? [N] > ');
            query(false);
            if yes then
              begin
                word:=entry^[i];
                kill_word;
              end; {if yes}
          end; {if found}
      end; {if suffix1}
    if length(word)>2 then
      begin
        suffix2:=copy(entry^[i],length(entry^[i])-1,2);
        if (suffix2='es') or (suffix2='ed') then
          begin
            word:=copy(entry^[i],1,length(entry^[i])-2);
            bsearch;
            if found then
              begin
                writeln;writeln('  ',word,'  ',entry^[i]);
                writeln;write('  Remove (Y/N ? [N] > ');
                query(false);
                if yes then
                  begin
                    word:=entry^[i];
                    kill_word;
                  end;
              end;
          end;

      end;


  end; {for i}
  waitkey;
end;

procedure prune;
var
  i : integer;
begin
  clrscr;
  writeln('* Edit Dictionary');
  writeln;write('  Begin Editing with Letter(s)  (def.="a") > ');
  readln(word);
  clrscr;
  bsearch;
  i:=word_no-1;
  repeat
    begin
      i:=i+1;
      writeln;writeln('> ',entry^[i]);
      writeln;write('  Delete (Y/N)? <N> ');
      query(false);
      if yes then
        begin
          word:=entry^[i];
          kill_word;
        end;
   writeln;write('  Continue (Y/N)? <Y> ');
      query(true);
    end;
  until not yes or (i=num_entries);
end;

procedure store;
begin
  writeln;write('* Storing Dictionary > ',num_entries,' entries');
  write_file;
end;

procedure abort;
begin
  writeln;writeln;write('* Aborting--Changes to Dictionary Not Stored ');writeln;
end;

procedure exit;
begin
  writeln;writeln;write('* Storing Dictionary > ',num_entries,' entries');
  write_file;writeln;
end;

procedure title;
begin
  clrscr;
  dashline;
  writeln(' SPELDX  Ver. ',version,'  Copr. 1985, 1987  M. Lee Murrah.  All Rights Reserved.');
  dashline;writeln;
  writeln('A dictionary maintenance program for SPELWELL.  Users are granted the right to');
  writeln('make and transfer copies so long as no consideration is charged.  Author');
  writeln('reserves exclusive right to prepare updates and derivative works for dist-');
  writeln('ribution.  Suggestions for changes and improvements, and information regarding');
  writeln('bugs should be directed to the author at 10 Cottage Grove Woods, SE, Cedar');
  writeln('Rapids, IA 52403, Tel: 319-365-6530.');
  writeln;writeln;write('* Loading Dictionary > ');
  read_file;writeln(num_entries,' entries');
  waitkey;
end;

begin
  title;
  repeat
    clrscr;
    dashline;
    writeln(' SPELDX  Ver. ',version,'  Copr. 1985, 1987  M. Lee Murrah.  All Rights Reserved.');
    dashline;writeln;
    writeln('                             <A> Add Word');
    writeln('                             <B> Browse Dictionary');
    writeln('                             <D> Delete Word');
    writeln('                             <E> Edit Dictionary');
    writeln('                             <F> Find Word');
    writeln('                             <M> Add Words from "Misspelled" File');
    writeln('                             <N> Number Entries');
    writeln('                             <R> Remove Redundant Words');
    writeln('                             <S> Store Dictionary');
    writeln;
    writeln('                             <Q> Quit--Discard Changes');
    writeln('                             <X> Exit--Save Changes');writeln;
    write('                             Your Choice > ');
    read(choice);writeln;
    choice:=upcase(choice);
  case choice of
  'A':add_entry;
  'B':display;
  'D':delete;
  'E':prune;
  'F':find;
  'M':add_list;
  'N':size_of_dx;
  'R':redundant;
  'Q':abort;
  'S':store;
  'X':exit;
  end;
  until choice in ['Q','X'];
  writeln;writeln('* Thanks for using SPELDX');writeln;
end.
